import os

from inspect import getmembers
from importlib import import_module
from itertools import chain

from zdict import dictionaries
from zdict.dictionary import DictBase


def get_dictionary_map():
    '''
    Auto discover dictionaries in package ``dictionaries``.
    Each dictionary class MUST be the subclass of ``DictBase``

    :return: a dict with {provider_name: cls}
        SomeDict.provider as key, the class as value
    '''
    package = 'zdict.dictionaries'
    exclude_files = ('template.py',)

    return {
        cls(None).provider: cls
        for _, cls in (
            chain.from_iterable(
                getmembers(mod, predicate=_is_dict)
                for mod in (
                    import_module(
                        '{}.{}'.format(package, f.partition('.py')[0]))
                    for f in os.listdir(dictionaries.__path__[0])
                    if (not f.startswith('_') and
                        f.endswith('.py') and
                        f not in exclude_files)
                )
            )
        )
    }


def _is_dict(cls):
    try:
        return issubclass(cls, DictBase) and not (cls is DictBase)
    except:
        return False



import json

from bs4 import BeautifulSoup

from zdict.dictionary import DictBase
from zdict.exceptions import NotFoundError
from zdict.models import Record


# [TODO]
#
# * let user choose en <-> spanish
# * some word's webpage use different CSS class ... (e.g. yo)
# * make code much more readable


class SpanishDict(DictBase):
    '''
    Tested words : ('soy', 'manzana', 'python', 'perdón')
    '''

    API = 'http://www.spanishdict.com/translate/{word}'

    @property
    def provider(self):
        return 'spanish'

    @property
    def title(self):
        return 'SpanishDict'

    def _get_url(self, word) -> str:
        return self.API.format(word=word)

    def show(self, record: Record):
        content = json.loads(record.content)

        self.color.print(content['word'], 'yellow')

        explains = content.get('explains')

        for data in explains:

            self.color.print(data[0], 'lwhite')     # speech

            for explain in data[1]:
                self.color.print(explain[0], 'lred')    # category
                for sentence in explain[1]:
                    self.color.print(
                        '{text}'.format(text=sentence[0]),
                        'org',
                        indent=2
                    )
                    if len(sentence) > 2:
                        self.color.print(
                            '{text}'.format(text=sentence[1]),
                            'lindigo',
                            indent=4
                        )
                        self.color.print(
                            '{text}'.format(text=sentence[2]),
                            'indigo',
                            indent=4
                        )
        print()

    def query(self, word: str):
        webpage = self._get_raw(word)
        data = BeautifulSoup(webpage, "html.parser")
        content = {}

        # word can be existing in both English & Spanish
        # choose English to Spanish here
        # 'translat-en' for English to Spanish
        # 'translat-es' for Spanish to English
        translate_en = data.find('div', id='translate-es')
        if translate_en:
            data = translate_en

        card = data.find('div', attrs={'class': 'card'})
        entry = card.find(
            # just get the first one
            attrs={'class': 'dictionary-entry'}
        )

        if not entry:
            raise NotFoundError(word)

        content['explains'] = []

        word_element = card.find(attrs={'class': 'source-text'})
        if word_element:
            content['word'] = word_element.text

        pattern1 = {'class': 'dictionary-neodict-indent-1'}
        pattern2 = {'class': 'dictionary-neodict-indent-2'}
        pattern3 = {'class': 'dictionary-neodict-indent-3'}
        pattern_order = {'class': 'dictionary-neodict-translation'}
        pattern_example = {'class': 'dictionary-neodict-example'}
        pattern1_en = {'class': 'dictionary-neoharrap-indent-1'}
        pattern2_en = {'class': 'dictionary-neoharrap-indent-2'}
        pattern_order_en = {'class': 'dictionary-neoharrap-translation'}

        speeches = card.find_all(attrs={'class': 'part_of_speech'})

        for (speech, category) in zip(
            speeches,
            entry.find_all(attrs=pattern1) or entry.find_all(attrs=pattern1_en)
        ):
            result = []
            content['explains'].append([speech.text, result])
            context = category.find(attrs={'class': 'context'}).text
            explains = []

            for explain in (category.find_all(attrs=pattern2) or
                            category.find_all(attrs=pattern2_en)):

                orders = (explain.find_all(attrs=pattern_order) or
                          explain.find_all(attrs=pattern_order_en))

                if orders:
                    # e.g.
                    #
                    #   ('a. forgiveness', 'b. pardon (law)')
                    #
                    indices = tuple(
                        map(
                            lambda x: x.text.replace('\xa0', ' ').strip(),
                            orders
                        )
                    )
                else:
                    continue

                examples = explain.find_all(attrs=pattern3)

                for (example, index) in zip(examples, indices):
                    t = tuple(example.find(attrs=pattern_example))
                    (spanish, english) = (t[0].text, t[2].text)
                    explains.append((index, spanish, english))

                if (not examples) and (len(indices) > 0):
                    for index in indices:
                        explains.append((index,))

            result.append([context, explains])

        record = Record(
            word=word,
            content=json.dumps(content),
            source=self.provider,
        )

        return record



from pytest import raises
from unittest.mock import Mock, patch

from zdict.dictionaries.moe import MoeDict
from zdict.exceptions import NotFoundError, QueryError
from zdict.models import Record
from zdict.zdict import get_args


class TestMoeDict:
    def setup_method(self, method):
        self.dict = MoeDict(get_args())

    def teardown_method(self, method):
        del self.dict

    def test__get_url(self):
        assert 'https://www.moedict.tw/uni/萌' == self.dict._get_url('萌')

    def test_provider(self):
        assert self.dict.provider == 'moe'

    def test_query_timeout(self):
        self.dict._get_raw = Mock(side_effect=QueryError('萌', 404))

        with raises(NotFoundError):
            self.dict.query('萌')

        self.dict._get_raw.assert_called_with('萌')

    @patch('zdict.dictionaries.moe.Record')
    def test_query_normal(self, Record):
        self.dict._get_raw = Mock(return_value='{}')
        self.dict.query('萌')
        Record.assert_called_with(word='萌', content='{}', source='moe')

    def test_show(self):
        content = '''
        {
            "heteronyms": [{
                "bopomofo": "ㄧㄢˋ",
                "bopomofo2": "yàn",
                "definitions": [{
                    "def": "假的、偽造的。",
                    "example": ["如：「贗品」。"],
                    "quote": ["..."],
                    "type": "形",
                    "synonyms": "尛",
                    "antonyms": "萌"
                }],
                "pinyin": "yàn"
            }],
            "non_radical_stroke_count": 15,
            "radical": "貝",
            "stroke_count": 22,
            "title": "贗"
        }
        '''
        r = Record(word='贗', content=content, source=self.dict.provider)

        # god bless this method, hope that it do not raise any exception
        self.dict.show(r)



from zdict.models import Record, db


class DictCompleter:
    def __init__(self):
        self.db = db
        self.db.connect()

    def __del__(self):
        self.db.close()

    def complete(self, text, state):
        if state == 0:  # new query
            self.records = iter(
                Record.select().where(Record.word.startswith(text)))

        return next(self.records).word



import peewee

from zdict.constants import DB_FILE


db = peewee.SqliteDatabase(DB_FILE)


class Record(peewee.Model):
    '''
    A model for storing the query results into the SQLite db.

    :param word: the vocabulary
    :param content: the query result of the vocabulary.
        It's a json document has the following spec.
        {
            'word': word,
            // storing the querying result.
            'pronounce': [
                ('key', 'value'),
                ...
            ],
            'sound': [
                ('type', 'url'),
                ...
                // type: (mp3|ogg)
            ],
            'explain': [
                ('speech',
                    (
                        'meaning',
                        ('sentence1', 'translation'),
                        ...
                    ),
                    ...
                ),
                ...
            ]
        }
    :param source: source of the content. May be Yahoo!, Google, ... Dict
    '''

    word = peewee.TextField()
    content = peewee.TextField()
    source = peewee.CharField()

    class Meta:
        database = db
        primary_key = peewee.CompositeKey('word', 'source')



import abc
import json

import requests

from zdict import exceptions
from zdict.models import Record, db
from zdict.utils import Color


class DictBase(metaclass=abc.ABCMeta):

    REQUIRED_TABLE = (
        Record,
    )

    def __init__(self, args):
        self.args = args
        self.db = db
        self.db.connect()

        for req in self.REQUIRED_TABLE:
            if not req.table_exists():
                req.create_table()

        self.color = Color()

    def __del__(self):
        del self.args
        self.db.close()
        del self.db

    @property
    @abc.abstractmethod
    def provider(self):
        '''
        Return the provider of online dictionary,
        this value is considered as `source` field in Record model.
        '''
        ...

    @property
    @abc.abstractmethod
    def title(self):
        '''
        Return the brief title of online dictionary,
        this value is considered as `source` field in Record model.
        '''
        ...

    @abc.abstractmethod
    def _get_url(self, word: str) -> str:
        '''
        Return the result of the current dict web url for the searching word.
        '''
        ...

    @abc.abstractmethod
    def show(self, record: Record):
        '''
        Define how to render the result of the specific dictionary.
        '''
        ...

    @abc.abstractmethod
    def query(self, word: str) -> Record:
        '''
        Define how to get the information from specific dictionary.
        Should return a record contains word, content and source.
        '''
        ...

    def query_db_cache(self, word: str) -> Record:
        try:
            record = Record.get(word=word, source=self.provider)
        except Record.DoesNotExist:
            return None
        else:
            return record

    def save(self, query_record: Record, word: str):
        db_record = self.query_db_cache(word)

        if db_record is None:
            query_record.save(force_insert=True)
        else:
            db_content = json.loads(db_record.content)
            query_content = json.loads(query_record.content)

            if db_content != query_content:
                db_record.content = query_record.content
                db_record.save()

    def show_provider(self):
        self.color.print('[' + self.provider + ']', 'blue')

    def show_url(self, word):
        self.color.print('(' + self._get_url(word) + ')', 'blue')

    def lookup(self, word):
        '''
        Main workflow for searching a word.
        '''

        word = word.lower()

        if self.args.show_provider:
            self.show_provider()

        if self.args.show_url:
            self.show_url(word)

        if not self.args.disable_db_cache:
            record = self.query_db_cache(word)

            if record:
                self.show(record)
                return

        try:
            record = self.query(word)
        except exceptions.NoNetworkError as e:
            self.color.print(e, 'red')
            print()
        except exceptions.TimeoutError as e:
            self.color.print(e, 'red')
            print()
        except exceptions.NotFoundError as e:
            self.color.print(e, 'yellow')
            print()
        else:
            self.save(record, word)
            self.show(record)
            return

    def _get_raw(self, word: str) -> str:
        '''
        Get raw data from http request

        :param word: single word
        '''

        try:
            res = requests.get(
                self._get_url(word), timeout=self.args.query_timeout
            )
        except requests.exceptions.ReadTimeout as e:
            raise exceptions.TimeoutError()
        except requests.exceptions.ConnectionError as e:
            errors = {
                "BlockingIOError(36, 'Operation now in progress')":
                    "exceptions.TimeoutError()",
                "Failed to establish a new connection":
                    "exceptions.NoNetworkError()",
            }
            for error, exception in errors.items():
                if error in str(e.args):
                    raise eval(exception)
            else:
                raise exceptions.UnexpectedError()
        except:
            raise exceptions.UnexpectedError()

        if res.status_code != 200:
            raise exceptions.QueryError(word, res.status_code)

        if self.args.debug:
            from bs4 import BeautifulSoup
            soup = BeautifulSoup(res.text, 'html.parser')
            print('='*30 + ' Start of debug info ' + '='*30)
            print(soup.prettify())
            print('='*30 + ' End of debug info ' + '='*30)

        return res.text



import json

from zdict.dictionary import DictBase
from zdict.exceptions import QueryError, NotFoundError
from zdict.models import Record


class MoeDict(DictBase):

    API = 'https://www.moedict.tw/uni/{word}'

    @property
    def provider(self):
        return 'moe'

    @property
    def title(self):
        return '萌典'

    def _get_url(self, word) -> str:
        return self.API.format(word=word)

    def show(self, record: Record):
        content = json.loads(record.content)

        # print word
        self.color.print(content.get('title', ''), 'yellow')

        for word in content.get('heteronyms', ''):

            # print pronounce
            for key, display in (
                ('bopomofo', '注音'),
                ('bopomofo2', '注音二式'),
                ('pinyin', '漢語拼音')
            ):
                self.color.print(display, end='')
                self.color.print(
                    '[' + word.get(key, '') + ']',
                    'lwhite',
                    end=' ',
                )
            print()
            print()

            # print explain
            for count, explain in enumerate(word.get('definitions', '')):

                self.color.print(
                    '{order}. {text}'.format(
                        order=count+1,
                        text=explain.get('def', '')
                    ),
                )

                if explain.get('synonyms'):
                    self.color.print(
                        '同義詞: {text}'.format(text=explain['synonyms']),
                        'magenta',
                        indent=2,
                    )

                if explain.get('antonyms'):
                    self.color.print(
                        '反義詞: {text}'.format(text=explain['antonyms']),
                        'magenta',
                        indent=2,
                    )

                for example in explain.get('example', ''):
                    self.color.print(
                        example,
                        'indigo',
                        indent=2,
                    )

                for quote in explain.get('quote', ''):
                    self.color.print(
                        '[引用] {text}'.format(text=quote),
                        'green',
                        indent=2,
                    )

                print()

    def query(self, word: str):
        try:
            content = self._get_raw(word)
        except QueryError as exception:
            raise NotFoundError(exception.word)

        record = Record(
                    word=word,
                    content=content,
                    source=self.provider,
                 )

        return record



from zdict.dictionaries.yahoo import YahooDict
from zdict.zdict import get_args


class TestyDict:
    def setup_method(self, method):
        self.dict = YahooDict(get_args())

    def teardown_method(self, method):
        del self.dict

    def test__get_url(self):
        url = 'https://tw.dictionary.search.yahoo.com/search?p=test'
        assert url == self.dict._get_url('test')



import itertools
import json
import re

from bs4 import BeautifulSoup

from zdict.dictionary import DictBase
from zdict.exceptions import NotFoundError
from zdict.models import Record


class YahooDict(DictBase):

    API = 'https://tw.dictionary.search.yahoo.com/search?p={word}'

    @property
    def provider(self):
        return 'yahoo'

    @property
    def title(self):
        return 'Yahoo Dictionary'

    def _get_url(self, word) -> str:
        return self.API.format(word=word)

    def show(self, record: Record):
        content = json.loads(record.content)

        # print word
        self.color.print(content['word'], 'yellow')

        # print pronounce
        for k, v in content.get('pronounce', []):
            self.color.print(k, end='')
            self.color.print(v, 'lwhite', end=' ')
        print()

        # print explain
        main_explanations = content.get('explain', [])
        if self.args.verbose:
            main_explanations.extend(content.get('verbose', []))

        for speech in main_explanations:
            self.color.print(speech[0], 'lred')
            for meaning in speech[1:]:
                self.color.print(
                    '{text}'.format(text=meaning[0]),
                    'org',
                    indent=2
                )
                for sentence in meaning[1:]:
                    if sentence:
                        print(' ' * 4, end='')
                        for i, s in enumerate(sentence.split('*')):
                            self.color.print(
                                s,
                                'lindigo' if i % 2 else 'indigo',
                                end=''
                            )
                        print()
        print()

    def query(self, word: str):
        webpage = self._get_raw(word)
        data = BeautifulSoup(webpage, "html.parser")
        content = {}

        # handle record.word
        try:
            content['word'] = data.find('span', id='term').text
        except AttributeError:
            raise NotFoundError(word)

        # handle pronounce
        pronu_value = data.find('span', id='pronunciation_pos').text
        if pronu_value:
            content['pronounce'] = []
            for match in re.finditer('(\w+)(\[.*?\])', pronu_value):
                content['pronounce'].append(match.group(1, 2))

        # handle sound
        pronu_sound = data.find(class_='proun_sound')
        if pronu_sound:
            content['sound'] = [
                ('mp3', pronu_sound.find(
                        class_='source',
                        attrs={'data-type': 'audio/mpeg'}
                    ).attrs['data-src']),
                ('ogg', pronu_sound.find(
                        class_='source',
                        attrs={'data-type': 'audio/ogg'}
                    ).attrs['data-src']),
            ]

        # Handle explain
        main_explanations = data.find(
            class_='dd algo explain mt-20 lst DictionaryResults'
        )
        if main_explanations:
            main_explanations = itertools.zip_longest(
                main_explanations.find_all(class_='compTitle mb-10'),
                main_explanations.find_all(
                    class_='compArticleList mb-15 ml-10',
                )
            )
        else:
            main_explanations = ""

        content['explain'] = []
        for part_of_speech, meaning in main_explanations:
            node = [part_of_speech.text] if part_of_speech else ['']

            for item in meaning.find_all('li', class_='ov-a'):
                pack = [item.find('h4').text]

                for example in (
                    tag for tag in item.find_all('span')
                    if 'line-height: 17px;' not in tag['style']
                ):
                    sentence = ''

                    for w in example.contents:
                        if w.name == 'b':
                            sentence += '*' + w.text + '*'
                        else:
                            try:
                                sentence += w
                            except:
                                pass

                    pack.append((sentence.strip()))
                node.append(pack)
            content['explain'].append(node)

            # verbose info
            part_of_speech_list, meaning_list = [], []
            content['verbose'] = []

            variation_explanations = data.find(
                class_='dd algo variation fst DictionaryResults'
            )
            if variation_explanations:
                part_of_speech_list.extend(
                    variation_explanations.find_all(class_='compTitle')
                )
                meaning_list.extend(
                    variation_explanations.find_all(class_='compArticleList')
                )

            additional_explanations = data.find(
                class_='dd algo othersNew lst DictionaryResults'
            )
            if additional_explanations:
                part_of_speech_list.extend(
                    additional_explanations.find_all(class_='compTitle mt-26')
                )
                meaning_list.extend(
                    additional_explanations.find_all(class_='compArticleList')
                )

            more_explanations = itertools.zip_longest(
                part_of_speech_list, meaning_list
            )

            for part_of_speech, meaning in more_explanations:
                node = [part_of_speech.text] if part_of_speech else ['']

                if meaning:
                    for item in meaning.find_all('li', class_='ov-a'):
                        pack = [item.find('h4').text]

                        for example in (
                            tag for tag in item.find_all('span')
                            if 'line-height: 17px;' not in tag['style']
                        ):
                            sentence = ''

                            for w in example.contents:
                                if w.name == 'b':
                                    sentence += '*' + w.text + '*'
                                else:
                                    try:
                                        sentence += w
                                    except:
                                        pass

                            pack.append((sentence.strip()))
                        node.append(pack)
                content['verbose'].append(node)

        record = Record(
                    word=word,
                    content=json.dumps(content),
                    source=self.provider,
                 )
        return record



import sys
import os

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand


def get_zdict_version():
    constants_file_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        'zdict/constants.py'
    )
    with open(constants_file_path) as constants:
        for line in constants:
            if line.startswith('VERSION'):
                code = compile(line, '<string>', 'single')
                version = code.co_consts[0]

        return version


def get_test_req():
    reqs = ['coverage', 'pytest', 'pytest-cov', 'pyjokes', 'flake8',
            'pytest-flake8']

    if not sys.platform.startswith('freebsd'):
        reqs.append('gnureadline')

    return reqs

version = get_zdict_version()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

requires = [
    'beautifulsoup4',
    'peewee',
    'requests',
]

if sys.platform == 'darwin':
    requires.append('gnureadline')


setup(
    packages=find_packages(exclude=['scripts']),
    scripts=['scripts/zdict'],
    install_requires=requires,
    tests_require=get_test_req(),
    cmdclass={'test': PyTest},

    name='zdict',
    version=version,
    author='Shun-Yi Jheng',
    author_email='M157q.tw@gmail.com',
    maintainer='Shun-Yi Jheng, Iblis Lin, Chang-Yen Chih, Chiu-Hsiang Hsu',
    maintainer_email=('M157q.tw@gmail.com,'
                      'e196819@hotmail.com,'
                      'michael66230@gmail.com,'
                      'wdv4758h@gmail.com'),
    url='https://github.com/zdict/zdict',
    keywords="cli, dictionary, framework",
    description="The last dictionary framework you need. (?)",
    long_description=open("README.rst").read(),
    download_url="https://github.com/zdict/zdict/archive/v{}.zip".format(
        version
    ),
    platforms=['Linux', 'Mac'],
    license="GPL3",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: Chinese (Traditional)",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Utilities",
    ],
)



from pytest import raises
from unittest.mock import Mock, patch

from zdict.dictionaries.spanish import SpanishDict
from zdict.exceptions import NotFoundError
from zdict.zdict import get_args


class TestSpanishDict:
    @classmethod
    def setup_class(cls):
        cls.dict = SpanishDict(get_args())
        cls.word = 'Spanish'
        cls.record = cls.dict.query(cls.word)

    @classmethod
    def teardown_class(cls):
        del cls.dict
        del cls.word
        del cls.record

    def test_provider(self):
        assert self.dict.provider == 'spanish'

    def test_title(self):
        assert self.dict.title == 'SpanishDict'

    def test__get_url(self):
        url = 'http://www.spanishdict.com/translate/{}'.format(self.word)
        assert url == self.dict._get_url(self.word)

    def test_show(self):
        # god bless this method, hope that it do not raise any exception
        self.dict.show(self.record)

    @patch('zdict.dictionaries.spanish.Record')
    def test_query_normal(self, Record):
        self.dict.query(self.word)
        Record.assert_called_with(
            word=self.word,
            content=self.record.content,
            source='spanish',
        )

    def test_query_not_found(self):
        self.dict._get_raw = Mock(return_value='<div class="card"><div/>')
        with raises(NotFoundError):
            self.dict.query(self.word)
        self.dict._get_raw.assert_called_with(self.word)
